



'''
##############################################################################################

Importing the required libraries 

##############################################################################################
'''

import torch
import warnings
warnings.filterwarnings('ignore')
import itertools
import time 
import torch.nn as nn
import torch.optim as optim
import itertools
import pdb

class Regressor:
    
    def __init__(self,x,y,names,dimension=None,sis_features=10,device='cpu',use_gpu=False):

        '''
        ###################################################################################################################

        x, y, names - are the outputs of the Feature Expansion class which defines the expanded feature space, target tensor, names of the expanded features to use in the equation

        dimension - defines the number of terms in the linear equation generation 

        sis_features - defines the number of top features needs to be considered for building the equation

        ###################################################################################################################
        '''
        self.device = device
        
        if use_gpu:
            
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            
        
        self.x = x.to(self.device)
        
        self.y = y.to(self.device)
        
        self.names = names
        
        if dimension !=None: 
            
            self.dimension = dimension
            
            self.sis_features = sis_features
            
        else: 
            self.dimension = 3 #Maximum terms we will be looking at...
            
            self.sis_features = 10
        

        # Transform the features into standardized format
        self.x_mean = self.x.mean(dim=0)
        
        self.x_std = self.x.std(dim=0)
        
        self.y_mean = self.y.mean()

        # Transform the target variable value to mean centered
        self.y_centered = self.y - self.y_mean
        
        self.x_standardized = ((self.x - self.x_mean)/self.x_std)

        self.scores = []
        
        self.indices = torch.arange(1, (self.dimension*self.sis_features+1)).view(self.dimension*self.sis_features,1).to(self.device)
        
        self.residual = torch.empty(self.y_centered.shape).to(self.device)
        
        self.x_std_clone = torch.clone(self.x_standardized)
        
        

    '''
    #######################################################################################################

    Constructs the linear equation based on the number of top sis features and the dimension requested.

    #######################################################################################################
    '''
    def higher_dimension(self,iteration):

        #Indices values that needs to be assinged zero 
        ind = (self.indices[:,-1][~torch.isnan(self.indices[:,-1])]).to(self.device)

        self.x_standardized[:,ind.tolist()] = 0

        scores= torch.abs(torch.mm(self.residual,self.x_standardized))

        scores[torch.isnan(scores)] = 0

        self.x_standardized[:,ind.tolist()] = self.x_std_clone[:,ind.tolist()]

        sorted_scores, sorted_indices = torch.topk(scores, k= self.sis_features)

        sorted_indices = sorted_indices.T
        
        sorted_indices_earlier = self.indices[:((iteration-1)*self.sis_features),(iteration-1)].unsqueeze(1)

        sorted_indices = torch.cat((sorted_indices_earlier,sorted_indices),dim=0)

        if sorted_indices.shape[0] < self.indices.shape[0]:
            
            remaining = (self.sis_features*self.dimension) - int(sorted_indices.shape[0])
            
            nan = torch.full((remaining,1),float('nan')).to(self.device)
            
            sorted_indices = torch.cat((sorted_indices,nan),dim=0)
            
            self.indices = torch.cat((self.indices,sorted_indices),dim=1)

        else:
            self.indices = torch.cat((self.indices,sorted_indices),dim=1)

        comb1 = self.indices[:,-1][~torch.isnan(self.indices[:,-1])]
        
        combinations_generated = torch.combinations(comb1,(int(self.indices.shape[1])-1))
        
        y_centered_clone = self.y_centered.unsqueeze(1).repeat(len(combinations_generated.tolist()),1,1).to(self.device)
        
        comb_tensor = self.x_standardized.T[combinations_generated.tolist(),:]
        
        x_p = comb_tensor.permute(0,2,1)
        

        sol, _, _, _ = torch.linalg.lstsq(x_p, y_centered_clone)
        
        predicted = torch.matmul(x_p,sol)
        
        residuals = y_centered_clone - predicted
        
        square = torch.square(residuals)
        
        mean = torch.mean(square,dim=1,keepdim=True)
        
        min_value, min_index = torch.min(mean, dim=0)
  
        coefs_min = torch.squeeze(sol[min_index]).unsqueeze(1)
        
        indices_min  = torch.squeeze(combinations_generated[min_index])
        
        non_std_coeff = ((coefs_min.T/self.x_std[indices_min.tolist()]))
        
        non_std_intercept = self.y.mean() - torch.dot(self.x_mean[indices_min.tolist()]/self.x_std[indices_min.tolist()],coefs_min.flatten())
        
        self.residual = self.y_centered - torch.mm(coefs_min.T,self.x_standardized[:,indices_min.tolist()].T)
        
        rmse = float(torch.sqrt(torch.mean(self.residual**2)))

        r2 = 1 - (float(torch.sum(self.residual**2))/float(torch.sum((self.y_centered)**2)))
        
        terms = []

        for i in range(len(non_std_coeff.squeeze())):
            
            ce = "{:.10f}".format(float(non_std_coeff.squeeze()[i]))
            
            term = str(ce) + "*" + str(self.names[int(indices_min[i])])
            
            terms.append(term)

        return float(rmse),terms,non_std_intercept,non_std_coeff,r2

    '''
    ##########################################################################################################################

    Defines the function to model the equation

    ##########################################################################################################################
    '''
    def regressor_fit(self):
        
        if self.x.shape[1] > self.sis_features*self.dimension :
            print('\n')
            
            print(f"************************************************ Starting sparse model building in {self.device} ************************************************ \n")
            
            print('\n')
            
        else:
            print('************************************************ Given Number of features in SIS screening is greater than the feature space created, changing the SIS features to shape of features created ************************************************ \n')
            
            self.sis_features = self.x.shape[1]
            
            self.indices = torch.arange(1, (self.dimension*self.sis_features+1)).view(self.dimension*self.sis_features,1).to(self.device)
            
        #Looping over the dimensions 
        final_equations = []
        for i in range(1,self.dimension+1):
            
            if i ==1:
                
                start_1D = time.time()

                #calculate the scores
                scores = torch.abs(torch.mm(self.y_centered.unsqueeze(1).T,self.x_standardized))

                #Set the NaN values claculation to zero, instead of removing 
                scores[torch.isnan(scores)] = 0

                #Sort the top number of scores based on the sis_features 
                sorted_scores, sorted_indices = torch.topk(scores,k=self.sis_features)
                
                sorted_indices = sorted_indices.T
                
                remaining = torch.tensor((self.sis_features*self.dimension) - int(sorted_indices.shape[0])).to(self.device)

                #replace the remaining indices with nan
                nan = torch.full((remaining,1),float('nan')).to(self.device)
                
                sorted_indices = torch.cat((sorted_indices,nan),dim=0)
                
                #store the sorted indices as next column
                self.indices = torch.cat((self.indices,sorted_indices),dim=1)
                
                selected_index = self.indices[0,1]
                
                x_in = self.x[:, int(selected_index)].unsqueeze(1)

                # Add a column of ones to x for the bias term
                x_with_bias = torch.cat((torch.ones_like(x_in), x_in), dim=1).to(self.device)

                #Calculate the intercept and coefficient, Non standardized
                coef1, _, _, _ = torch.linalg.lstsq(x_with_bias, self.y)

                #Calculate the residuals based on the standardized and centered values
                x_in1 = self.x_standardized[:, int(selected_index)].unsqueeze(1)

                # Add a column of ones to x for the bias term
                x_with_bias1 = torch.cat((torch.ones_like(x_in1), x_in1), dim=1)
                
                coef, _, _, _ = torch.linalg.lstsq(x_with_bias1, self.y_centered)

                self.residual = (self.y_centered - (coef[1]*self.x_standardized[:, int(selected_index)])).unsqueeze(1).T
                
                rmse = float(torch.sqrt(torch.mean(self.residual**2)))
                
                r2 = 1 - (float(torch.sum(self.residual**2))/float(torch.sum((self.y_centered)**2)))
                
                coefficient = coef[1]

                intercept = self.y.mean() - torch.dot((self.x_mean[int(selected_index)]/self.x_std[int(selected_index)]).reshape(-1), coef[1].reshape(-1))#coef1[0]

                if intercept > 0:
                    
                    coefficient = coef[1]/self.x_std[int(selected_index)]
                    
                    coefficient = "{:.6f}".format(float(coefficient))
                    
                    equation = str(float(coefficient)) + '*' + str(self.names[int(selected_index)]) + '+' + str(float(intercept))
                    
                    print('Equation: ', equation)
                    
                    print('\n')
                    
                    print('RMSE: ', rmse)
                    
                    print('R2::',r2)
                    
                else:
                    
                    coefficient = coef[1]/self.x_std[int(selected_index)]
                    
                    coefficient = "{:.6f}".format(float(coefficient))
                    
                    equation = str(float(coefficient)) + '*' + str(self.names[int(selected_index)])  + str(float(intercept))
                    
                    print('Equation: ', equation)
                    
                    print('\n')
                    
                    print('RMSE: ', rmse)
                    
                    print('R2::',r2)
                    
                print('\n')
                
                print('************************************************ Time taken to generate one dimensional equation: ', time.time()-start_1D,' seconds ************************************************ \n')
                
                print('\n')
                final_equations.append(equation)
                if self.device == 'cuda':torch.cuda.empty_cache()
                
                
                
            else:
                
                start = time.time()

                rmse,terms,intercept,coefs,r2 = self.higher_dimension(i)
                
                equation =''
                
                for k in range(len(terms)):
                    
                    print(f'{k+1} term in the equation is {terms[k]}')
                    print('\n')

                    if coefs.flatten()[k] > 0:
                        
                        equation = equation + ' + ' + (str(terms[k]))+'  '
                        
                    else:
                        
                        equation = equation + (str(terms[k])) + '  '

                if intercept>0: equation = equation +'+'+ str(float(intercept))
                
                else: equation = equation + str(float(intercept))

                print('Equation: ',equation[:len(equation)-1])
                print('\n')
                

                print('RMSE:',float(rmse))
                print('\n')
                
                print('R2::',r2)

                print(f'************************************************ Time taken for {i} dimension is: ', time.time()-start,' seconds ************************************************ \n')
                
                final_equations.append(equation)
                
                if self.device == 'cuda': torch.cuda.empty_cache()
                
                

        return float(rmse),equation,r2,final_equations




'''
##############################################################################################

Importing the required libraries

##############################################################################################
'''
import torch
import pandas as pd
import numpy as np
import warnings
import itertools
import time
from itertools import combinations
from . import combinations_construction

class feature_space_construction:

  '''
  ##############################################################################################################

  Define global variables like number of operators and the input data frame and the operator set given

  ##############################################################################################################
  '''
  def __init__(self,operators,df,no_of_operators=None,device='cpu',initial_screening=None,custom_unary_functions=None,custom_binary_functions=None):

    print(f'************************************************ Starting Feature Space Construction in {device} ************************************************ \n')
    print('\n')
    '''
    ###########################################################################################

    no_of_operators - defines the presence of operators (binary or unary) in the expanded features space

    For example: if no_of_operators = 2 then the space will be limited to formation of features with 3 operators (x1+x2)/x3 or exp(x1+x2)

    ###########################################################################################
    '''
    self.no_of_operators = no_of_operators

    self.df = df
    '''
    ###########################################################################################

    operators [list type]: Defines the mathematical operators needs to be used in the feature expansion

    Please look at the README.md for type of mathematical operators allowed

    ###########################################################################################
    '''
    self.operators = operators
    
    self.device = torch.device(device)

    # Filter the dataframe by removing the categorical datatypes and zero variance feature variables

    self.df = self.df.select_dtypes(include=['float64','int64'])

    # Compute the variance of each column
    variance = self.df.var()

    # Get the names of the zero variance columns
    zero_var_cols = variance[variance == 0].index

    # Drop the zero variance columns from the dataframe
    self.df = self.df.drop(zero_var_cols, axis=1)
    
    # Pop out the Targer variable of the problem and convert to tensor
    self.df.rename(columns = {f'{self.df.columns[0]}':'Target'},inplace=True)
    
    self.Target_column = torch.tensor(self.df.pop('Target')).to(self.device)
    
    
    if initial_screening != None:
        
        self.screening = initial_screening[0]
        
        self.quantile = initial_screening[1]
        
        self.df = self.feature_space_screening(self.df)
        


    # Create the feature values tensor
    self.df_feature_values = torch.tensor(self.df.values).to(self.device)
    
    self.columns = self.df.columns.tolist()

    #Create a dataframe for appending new datavalues
    self.new_features_values = pd.DataFrame()

    #Creating empty tensor and list for single operators (Unary operators)
    self.feature_values_unary = torch.empty(self.df.shape[0],0).to(self.device)
    
    self.feature_names_unary = []

    #creating empty tensor and list for combinations (Binary Operators)
    self.feature_values_binary = torch.empty(self.df.shape[0],0).to(self.device)
    
    self.feature_names_binary = []
    
    self.custom_unary_functions = custom_unary_functions
    
    self.custom_binary_functions = custom_binary_functions
    
    

  '''
  ###############################################################################################################

  Construct all the features that can be constructed using the single operators like log, exp, sqrt etc..

  ###############################################################################################################
  '''
  def feature_space_screening(self,df_sub):
        
        from sklearn.feature_selection import mutual_info_regression
        from scipy.stats import spearmanr

        if self.screening == 'spearman':
            spear = spearmanr(df_sub.to_numpy(),self.Target_column,axis=0)
            screen1 = abs(spear.statistic)
            if screen1.ndim>1:screen1 = screen1[:-1,-1]
        elif self.screening=='mi':
            screen1 = mutual_info_regression(df_sub.to_numpy(), self.Target_column.numpy())
            
        
        
        df_screening = pd.DataFrame()
        df_screening['Feature variables'] = df_sub.columns
        df_screening['screen1'] = screen1
        df_screening = df_screening.sort_values(by = 'screen1',ascending= False).reset_index(drop=True)
        quantile_screen=df_screening.screen1.quantile(self.quantile)
        
        filtered_df = df_screening[(df_screening.screen1 > quantile_screen)].reset_index(drop=True)
        
        if filtered_df.shape[0]==0:
            filtered_df = df_screening[:int(df_sub.shape[1]/2)]

        df_screening1 = df_sub.loc[:,filtered_df['Feature variables'].tolist()]
        
        
        return df_screening1
    
  def construct_function(self,functions_tuple, var):
      
    result = var  # Start with the input variable
    text1 = ''  # Initialize text1 as an empty string
    text2 = ''  # Initialize text2 as an empty string
    
    for func in functions_tuple:
        # Check for power function of the form pow(n)
        if func.startswith('pow'):
            # Extract the power value inside parentheses
            power_value = int(re.search(r'pow\((\d+)\)', func).group(1))
            result = result ** power_value  # Use ** for power operation
            text2 = f")**{power_value}" + text2
            text1 = "(" + text1
        # Check if the function has a negative sign
        elif func.startswith('-'):
            func_name = func[1:]  # Remove the negative sign
            result = -getattr(torch, func_name)(result)  # Apply the function with a negative sign
            text1 = f"-{getattr(torch, func_name).__name__}(" + text1
            text2 = ")" + text2
        else:
            result = getattr(torch, func)(result)  # Apply the function normally
            text1 = f"{getattr(torch, func).__name__}(" + text1
            text2 = ")" + text2

    return result, text1, text2  


  def single_variable(self,operators_set,i):


    #Looping over operators set to get the new features/predictor variables
    
    if len(operators_set)==0 and self.custom_unary_functions!=None:
        
        for func in self.custom_unary_functions:
            
            self.feature_values_11 = torch.empty(self.df.shape[0],0).to(self.device)
            
            feature_names_12 =[]
            
            self.feature_values_11,t1,t2 = self.construct_function(func,self.df_feature_values)
            
            feature_names_12.extend(list(map(lambda x: str(t1) +x + str(t2), self.columns)))
            
            self.feature_values_unary = torch.cat((self.feature_values_unary,self.feature_values_11),dim=1)
            
            self.feature_names_unary.extend(feature_names_12)

            del self.feature_values_11, feature_names_12

    for op in operators_set:

        self.feature_values_11 = torch.empty(self.df.shape[0],0).to(self.device)
        feature_names_12 =[]

        # Performs the exponential transformation of the given feature space
        if op == 'exp':
            exp = torch.exp(self.df_feature_values)
            self.feature_values_11 = torch.cat((self.feature_values_11,exp),dim=1)
            feature_names_12.extend(list(map(lambda x: '(exp('+ x + "))", self.columns)))

        # Performs the natural lograithmic transformation of the given feature space
        elif op =='ln':
            
            ln = torch.log(self.df_feature_values)
            self.feature_values_11 = torch.cat((self.feature_values_11,ln),dim=1)
            feature_names_12.extend(list(map(lambda x: '(ln('+x + "))", self.columns)))

        # Performs the lograithmic transformation of the given feature space
        elif op =='log':
            log10 = torch.log10(self.df_feature_values)
            self.feature_values_11 = torch.cat((self.feature_values_11,log10),dim=1)
            feature_names_12.extend(list(map(lambda x: '(log('+x + "))", self.columns)))

        # Performs the power transformations of the feature variables..
        
        elif "pow" in op:
            
            import re
            
            pattern = r'\(([^)]*)\)'
            matches = re.findall(pattern, op)
            op = eval(matches[0])
            
            transformation = torch.pow(self.df_feature_values,op)
            self.feature_values_11 = torch.cat((self.feature_values_11,transformation),dim=1)
            feature_names_12.extend(list(map(lambda x: '('+x + f")^{matches[0]}", self.columns)))

        # Performs the sine function transformation of the given feature space
        elif op =='sin':
            sin = torch.sin(self.df_feature_values)
            self.feature_values_11 = torch.cat((self.feature_values_11,sin),dim=1)
            feature_names_12.extend(list(map(lambda x: '(sin('+x + "))", self.columns)))

        # Performs the hyperbolic sine function transformation of the given feature space
        elif op =='sinh':
            sin = torch.sinh(self.df_feature_values)
            self.feature_values_11 = torch.cat((self.feature_values_11,sin),dim=1)
            feature_names_12.extend(list(map(lambda x: '(sin('+x + "))", self.columns)))

        # Performs the cosine transformation of the given feature space
        elif op =='cos':
            cos = torch.cos(self.df_feature_values)
            self.feature_values_11 = torch.cat((self.feature_values_11,cos),dim=1)
            feature_names_12.extend(list(map(lambda x: '(cos('+x + "))", self.columns)))

        # Performs the hyperbolic cosine transformation of the given feature space
        elif op =='cosh':
            cos = torch.cosh(self.df_feature_values)
            self.feature_values_11 = torch.cat((self.feature_values_11,cos),dim=1)
            feature_names_12.extend(list(map(lambda x: '(cos('+x + "))", self.columns)))

        # Performs the hyperbolic tan transformation of the given feature space
        elif op =='tanh':
            cos = torch.tanh(self.df_feature_values)
            self.feature_values_11 = torch.cat((self.feature_values_11,cos),dim=1)
            feature_names_12.extend(list(map(lambda x: '(cos('+x + "))", self.columns)))
            
        # Performs the Inverse transformation of the given feature space
        elif op =='^-1':
            reciprocal = torch.reciprocal(self.df_feature_values)
            self.feature_values_11 = torch.cat((self.feature_values_11,reciprocal),dim=1)
            feature_names_12.extend(list(map(lambda x: '(('+x + ")^(-1))", self.columns)))

        # Performs the Inverse exponential transformation of the given feature space
        elif op =='exp(-1)':
            exp = torch.exp(self.df_feature_values)
            expreciprocal = torch.reciprocal(exp)
            self.feature_values_11 = torch.cat((self.feature_values_11,expreciprocal),dim=1)
            feature_names_12.extend(list(map(lambda x: '(exp('+x + ")^(-1))", self.columns)))
            
        elif op =='+1':
            add1 = self.df_feature_values + 1
            self.feature_values_11 = torch.cat((self.feature_values_11,add1),dim=1)
            feature_names_12.extend(list(map(lambda x: '('+x + "+1)", self.columns)))
            
        elif op =='-1':
            sub1 = self.df_feature_values - 1
            self.feature_values_11 = torch.cat((self.feature_values_11,sub1),dim=1)
            feature_names_12.extend(list(map(lambda x: '('+x + "-1)", self.columns)))
            
        elif op =='/2':
            div2 = self.df_feature_values/2
            self.feature_values_11 = torch.cat((self.feature_values_11,div2),dim=1)
            feature_names_12.extend(list(map(lambda x: '('+x + "/2)", self.columns)))
            
        elif self.custom_unary_functions!=None:
            
            for func in self.custom_functions:
                
                
                self.feature_values_11,t1,t2 = self.construct_function(func,self.df_feature_values_11)
                
                feature_names_12.extend(list(map(lambda x: str(t1) +x + str(t2), self.columns)))
        

        self.feature_values_unary = torch.cat((self.feature_values_unary,self.feature_values_11),dim=1)
        
        self.feature_names_unary.extend(feature_names_12)

        del self.feature_values_11, feature_names_12



    #Check for empty lists
    if len(self.feature_names_unary) == 0:
        return self.feature_values_unary, self.feature_names_unary
    else:
        # create Boolean masks for NaN and Inf values

        nan_columns = torch.any(torch.isnan(self.feature_values_unary), dim=0)
        inf_columns = torch.any(torch.isinf(self.feature_values_unary), dim=0)
        nan_or_inf_columns = nan_columns | inf_columns

        # Remove columns from tensor
        self.feature_values_unary = self.feature_values_unary[:, ~nan_or_inf_columns]

        # Remove corresponding elements from list
        self.feature_names_unary = [elem for i, elem in enumerate(self.feature_names_unary) if not nan_or_inf_columns[i]]
        
        if self.no_of_operators!=None:
            
            if i+1!=self.no_of_operators:
                
                #Get the duplicate columns in the feature space created..
                unique_columns, indices = torch.unique(self.feature_values_unary,sorted=False, dim=1, return_inverse=True)
                
                # Get the indices of the unique columns
                unique_indices = indices.unique()
        
                # Remove duplicate columns
                self.feature_values_unary = self.feature_values_unary[:, unique_indices]
    
                # Remove the corresponding elements from the list of feature names..
                self.feature_names_unary = [self.feature_names_unary[i] for i in unique_indices.tolist()]
            
        
        return self.feature_values_unary, self.feature_names_unary



  '''
  ################################################################################################

  Defining method to perform the combinations of the variables with the initial feature set
  ################################################################################################
  '''
  def combinations(self,operators_set,i):
      
      
      if len(operators_set)==0 and self.custom_binary_functions!=None:
          
          self.feature_values11 = torch.empty(self.df.shape[0],0).to(self.device)
          feature_names_11 = []
          
          constructor = FeatureConstructor(self.df_feature_values, self.columns)
          
          results, expressions = constructor.construct_generic_features(self.custom_binary_functions)
          self.feature_values11 = torch.cat((self.feature_values11,results),dim=1)
          feature_names_11.extend(expressions)
          self.feature_values_binary = torch.cat((self.feature_values_binary,self.feature_values11),dim=1)
          self.feature_names_binary.extend(feature_names_11)
          del self.feature_values11,feature_names_11
          
          
          
      for op in operators_set:
          s = time.time()
          #getting list of cobinations without replacement using itertools
          combinations1 = list(combinations(self.columns,2))

          combinations2 = torch.combinations(torch.arange(self.df_feature_values.shape[1]),2)

          comb_tensor = self.df_feature_values.T[combinations2,:]

          #Reshaping to match
          x_p = comb_tensor.permute(0,2,1)

          del comb_tensor,combinations2 #Deleting to release the memory
          self.feature_values11 = torch.empty(self.df.shape[0],0).to(self.device)
          feature_names_11 = []
          
          

          # Performs the addition transformation of feature space with the combinations generated
          if op =='+':
              sum = torch.sum(x_p,dim=2).T
              self.feature_values11 = torch.cat((self.feature_values11,sum),dim=1)
              feature_names_11.extend(list(map(lambda comb: '('+'+'.join(comb)+')', combinations1)))
              

          # Performs the subtraction transformation of feature space with the combinations generated
          elif op =='-':
              sub = torch.sub(x_p[:,:,0],x_p[:,:,1]).T
              self.feature_values11 = torch.cat((self.feature_values11,sub),dim=1)
              feature_names_11.extend(list(map(lambda comb: '('+'-'.join(comb)+')', combinations1)))

          # Performs the division transformation of feature space with the combinations generated
          elif op == '/':
              div1 = torch.div(x_p[:,:,0],x_p[:,:,1]).T
              div2 = torch.div(x_p[:,:,1],x_p[:,:,0]).T
              self.feature_values11 = torch.cat((self.feature_values11,div1,div2),dim=1)
              feature_names_11.extend(list(map(lambda comb: '('+'/'.join(comb)+')', combinations1)))
              feature_names_11.extend(list(map(lambda comb: '('+'/'.join(comb[::-1])+')', combinations1)))

          # Performs the multiplication transformation of feature space with the combinations generated
          elif op == '*':
              mul = torch.multiply(x_p[:,:,0],x_p[:,:,1]).T
              self.feature_values11 = torch.cat((self.feature_values11,mul),dim=1)
              feature_names_11.extend(list(map(lambda comb: '('+'*'.join(comb)+')', combinations1)))
        
          elif self.custom_binary_functions!=None:
              
              constructor = FeatureConstructor(self.df_feature_values, self.columns)
              
              results, expressions = constructor.construct_generic_features(self.custom_binary_functions)
              self.feature_values11 = torch.cat((self.feature_values11,results),dim=1)
              feature_names_11.extend(expressions)
              
          self.feature_values_binary = torch.cat((self.feature_values_binary,self.feature_values11),dim=1)
          self.feature_names_binary.extend(feature_names_11)
          del self.feature_values11,feature_names_11
          #print('Operator::',op,time.time()-s)
      
      #Checking whether the lists are empty
      if len(self.feature_names_binary) == 0:
          return self.feature_values_binary, self.feature_names_binary
      
      else:
          #Removing Nan and inf columns from tenosr and corresponding variable name form the list

          nan_columns = torch.any(torch.isnan(self.feature_values_binary), dim=0)
          inf_columns = torch.any(torch.isinf(self.feature_values_binary), dim=0)
          nan_or_inf_columns = nan_columns | inf_columns

          # Remove columns from tensor
          self.feature_values_binary = self.feature_values_binary[:, ~nan_or_inf_columns]

          # Remove corresponding elements from list
          self.feature_names_binary = [elem for i, elem in enumerate(self.feature_names_binary) if not nan_or_inf_columns[i]]

          #Get the duplicate columns in the feature space created..
          if self.operators !=None:
              
              if i+1 != self.no_of_operators:
                  
                  unique_columns, indices = torch.unique(self.feature_values_binary, sorted=False,dim=1, return_inverse=True)
                  
                  # Get the indices of the unique columns
                  unique_indices = indices.unique()
        
                  # Remove duplicate columns
                  self.feature_values_binary = self.feature_values_binary[:, unique_indices]
                  
                  
                  # Remove the corresponding elements from the list of feature names..
                  self.feature_names_binary = [self.feature_names_binary[i] for i in unique_indices.tolist()]

          #Returning the featurespace created
          return self.feature_values_binary,self.feature_names_binary #created_space


  '''
  ##########################################################################################################

  Creating the space based on the given set of conditions

  ##########################################################################################################

  '''

  def feature_space(self):
 
        basic_operators = [op for op in self.operators if op in ['+','-','*','/']]
        other_operators = [op for op in self.operators if op not in ['+','-','*','/']]
        for i in range(1,self.no_of_operators):
            
            start_time = time.time()
            
            print(f'************************************************ Starting {i} level of feature expansion...************************************************ \n')
    
            #Performs the feature space expansion based on the binary operator set provided
            values, names = self.combinations(basic_operators,i)
        
            # Performs the feature space expansion based on the unary operator set provided
            values1, names1 = self.single_variable(other_operators,i)
        
            features_created = torch.cat((values,values1),dim=1)
            
            del values, values1
            
            names2 = names + names1
            
            del names,names1
            
            self.df_feature_values = torch.cat((self.df_feature_values,features_created),dim=1)
            
            self.columns.extend(names2)
            
            del features_created,names2
            
            print(f'************************************************ {i} Feature Expansion Completed with feature space size:::',self.df_feature_values.shape[1],'************************************************ \n')
            
            print('************************************************ Time taken to create the space is:::', time.time()-start_time, ' Seconds...************************************************ \n')
            
            
        return self.df_feature_values, self.Target_column,self.columns
    







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  4 00:56:39 2024

@author: muthyala.7
"""
import torch
import re

class FeatureConstructor:
    def __init__(self, df_feature_values, columns):
        
        self.columns = columns
        self.device = 'cpu'
        self.df_feature_values = df_feature_values

    def apply_function_tensor(self, tensor, func):
        
        if func.startswith('pow'):
            power_value = int(re.search(r'pow\((\d+)\)', func).group(1))
            return tensor ** power_value
        elif func == 'div':
            return torch.div(tensor, 1)  # Default division by 1 if no second argument
        elif func.startswith('-'):
            func_name = func[1:]
            return -getattr(torch, func_name)(tensor)
        else:
            return getattr(torch, func)(tensor)

    def apply_operation(self, tensor1, tensor2, op):
        # Apply operation defined in torch (op is now passed directly from feature_def)
        return op(tensor1, tensor2)

    def construct_generic_features(self, feature_definition):
        n_features = len(self.columns)

        # Create a grid of combinations
        x, y = torch.meshgrid(torch.arange(n_features), torch.arange(n_features), indexing='ij')

        # Stack and reshape the results
        combinations_indices = torch.stack((x.flatten(), y.flatten()), dim=1)

        # Create a 3D tensor of shape (n_combinations, n_samples, 2)
        comb_tensor = self.df_feature_values[:, combinations_indices].permute(1, 0, 2)

        results = []
        expressions = []

        for part in feature_definition:
            var1_idx = 0
            var2_idx = 1
            
            # Ensure the indices are within bounds of the 3D tensor's last dimension
            if var1_idx >= 2 or var2_idx >= 2:
                raise IndexError(f"var1_idx or var2_idx is out of bounds. Maximum index allowed is 1, got var1_idx={var1_idx}, var2_idx={var2_idx}")
            
            functions1 = part['functions1']
            functions2 = part['functions2']
            op = part['operation']  # Operation is now part of the feature definition
            final_functions = part.get('final_functions', [])  # Get final functions if they exist

            var1 = comb_tensor[:, :, var1_idx]
            var2 = comb_tensor[:, :, var2_idx]

            # Apply functions to the first variable only if there are any functions specified
            if functions1:
                for func in functions1:
                    var1 = self.apply_function_tensor(var1, func)

            # Apply functions to the second variable only if there are any functions specified
            if functions2:
                for func in functions2:
                    var2 = self.apply_function_tensor(var2, func)

            # Apply the operation between the two variables
            op_result = self.apply_operation(var1, var2, op)

            # Apply final functions on the operation result
            if len(final_functions)>=1:
                for func in final_functions:
                    op_result = self.apply_function_tensor(op_result, func)

            # Create expression strings
            expr1_template = [self.columns[idx.item()] for idx in combinations_indices[:, var1_idx]]
            expr2_template = [self.columns[idx.item()] for idx in combinations_indices[:, var2_idx]]

            # Apply functions to expression strings
            for func in functions1:
                if func.startswith('pow'):
                    power_value = re.search(r'pow\((\d+)\)', func).group(1)
                    expr1_template = [f"({expr})**{power_value}" for expr in expr1_template]
                elif func.startswith('-'):
                    expr1_template = [f"-{expr}" for expr in expr1_template]
                else:
                    expr1_template = [f"{func}({expr})" for expr in expr1_template]

            for func in functions2:
                if func.startswith('pow'):
                    power_value = re.search(r'pow\((\d+)\)', func).group(1)
                    expr2_template = [f"({expr})**{power_value}" for expr in expr2_template]
                elif func.startswith('-'):
                    expr2_template = [f"-{expr}" for expr in expr2_template]
                else:
                    expr2_template = [f"{func}({expr})" for expr in expr2_template]

            # Combine expressions with the operation
            combined_expr = [f"({expr1}) {op.__name__} ({expr2})" for expr1, expr2 in zip(expr1_template, expr2_template)]
            
            # Combine final expressions only if final functions are specified
            final_expr = combined_expr
            
            
            if final_functions:
                final_expr = []
                final_expr = [f"{func}(({expr1}) {op.__name__} ({expr2}))" for expr1, expr2 in zip(expr1_template, expr2_template) for func in final_functions]


            
            results.append(op_result)
            expressions.extend(final_expr)

        return torch.cat(results, dim=1).T, expressions






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 17 09:22:50 2023

@author: muthyala.7
"""

from .import FeatureSpaceConstruction as fc
from .import DimensionalFeatureSpaceConstruction as dfc
from .Regressor import Regressor
from .Regressor_dimension import Regressor


import sys
import time
import pdb
import numpy as np 
import pandas as pd 
import time
from sympy import symbols
from sklearn.model_selection import train_test_split





class SissoModel:

  def __init__(self,data,operators=None,multi_task = None,n_expansion=None,n_term=None,k=20,device='cpu',use_gpu = False,
               relational_units = None,initial_screening = None,dimensionality=None,output_dim = None,
               regressor_screening = None,custom_unary_functions = None ,custom_binary_functions=None):

    self.operators = operators
    
    if self.operators == None: sys.exit('!! Please provide the valid operator set!!')
    
    self.df=data

    self.no_of_operators = n_expansion
    
    if self.no_of_operators == None: 
        
        print('!! Warning:: Number of feature expansions is not provided, considering default configuration!! \n')
        
        self.no_of_operators = 3
        
    self.device = device
    
    if n_term == None: 
        
        print('Warning:: Number of terms in equation is not provided, considering default configuration..!! \n')
        
        self.dimension = 3
        
    else: self.dimension = n_term
        
    else: self.sis_features = k
    
    self.relational_units = relational_units
    
    self.initial_screening = initial_screening
    
    self.dimensionality = dimensionality
    
    self.output_dim = output_dim
    
    self.regressor_screening = regressor_screening
    
    self.use_gpu = use_gpu
    
    self.custom_unary_functions = custom_unary_functions
    
    self.custom_binary_functions = custom_binary_functions
    
    self.multi_task = multi_task
    
    if multi_task!=None:
        
        self.multi_task_target = multi_task[0]
        
        self.multi_task_features = multi_task[1]
    

      
  def fit(self):
      
    if self.dimensionality == None:
        
        if self.operators==None: sys.exit('Please provide the operators set for the non dimensional Regression!!')
        
        if self.multi_task!=None:
            
            print('Performing MultiTask Symbolic regression!!..')
            
            equations,rmses,r2s =[],[],[]
            
            for i in range(len(self.multi_task_target)):
                
                #Get the target variable and feature variabls 
                print('Performing symbolic regression of',i+1,'Target variables....')
                
                list1 =[]
                
                list1.extend([self.multi_task_target[i]]+self.multi_task_features[i])
                
                df1 = self.df.iloc[:,list1]
                
                
                x,y,names = fc.feature_space_construction(self.operators,df1,self.no_of_operators,self.device,self.initial_screening,self.custom_unary_functions,self.custom_binary_functions).feature_space()
                    
                from .Regressor import Regressor
                  
                rmse, equation,r2,final_eq =  Regressor(x,y,names,self.dimension,self.sis_features,self.device,self.use_gpu).regressor_fit()
                
                equations.append(final_eq)
                
                rmses.append(rmse)
                
                r2s.append(r2)
                    
                if i+1 == len(self.multi_task_target):
                    
                    print('Equations found::',equations)
                    
                    return rmses, equations, r2s
                
                else: continue
                
        
        else:
            
            x,y,names = fc.feature_space_construction(self.operators,self.df,self.no_of_operators,self.device,self.initial_screening,self.custom_unary_functions,self.custom_binary_functions).feature_space()
            
            from .Regressor import Regressor
            
            rmse, equation,r2,final_eq =  Regressor(x,y,names,self.dimension,self.sis_features,self.device,self.use_gpu).regressor_fit()
        
            return rmse, equation, r2,final_eq
  
    else: 
        
        if self.multi_task!=None:
            
            print('************************************************ Performing MultiTask Symbolic regression!!..************************************************ \n')
            
            equations,rmses,r2s =[],[],[]
            
            for i in range(len(self.multi_task_target)):
                
                #Get the target variable and feature variabls 
                print('************************************************ Performing symbolic regression of',i+1,'Target variables....************************************************ \n')
                
                list1 =[]
                
                list1.extend([self.multi_task_target[i]]+self.multi_task_features[i])
                
                df1 = self.df.iloc[:,list1]
                
                
                    
                x,y,names,dim = dfc.feature_space_construction(self.df,self.operators,self.relational_units,self.initial_screening,self.no_of_operators,self.device,self.dimensionality).feature_expansion()
                    
                    #print(names)
                from .Regressor_dimension import Regressor
                    
                rmse,equation,r2 = Regressor(x,y,names,dim,self.dimension,self.sis_features,self.device,self.output_dim,self.regressor_screening,self.use_gpu).regressor_fit()
                    
                equations.append(equation)
                
                rmses.append(rmse)
                
                r2s.append(r2)
                    
                    
                if i+1 == len(self.multi_task_target):
                        
                    print('Equations found::',equations)
                        
                    return rmses, equations, r2s
                    
                else: continue
                
        
        
        else:
            
            x,y,names,dim = dfc.feature_space_construction(self.df,self.operators,self.relational_units,self.initial_screening,self.no_of_operators,self.device,self.dimensionality).feature_expansion()
            
            #print(names)
            from .Regressor_dimension import Regressor
            
            rmse,equation,r2 = Regressor(x,y,names,dim,self.dimension,self.sis_features,self.device,self.output_dim,self.regressor_screening,self.use_gpu).regressor_fit()
            
            return rmse,equation,r2





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 11 12:00:15 2024

@author: muthyala.7
"""

import pandas as pd 
import numpy as np 
import torch 
import time 
from scipy.stats import spearmanr
import sys
import pdb


class Regressor:
    
    def __init__(self,x,y,names,dimensionality,dimension=None,sis_features=10,device='cpu',output_dim=None,screening=None,use_gpu=False):

        '''
        ###################################################################################################################

        x, y, names - are the outputs of the Feature Expansion class which defines the expanded feature space, target tensor, names of the expanded features to use in the equation
        
        dimensionality -- describes the units of the variables created.. 
        
        dimension - defines the number of terms in the linear equation generation 
        
        output_dim - dimensionality of the target variable

        sis_features - defines the number of top features needs to be considered for building the equation

        ###################################################################################################################
        '''
        self.device = device
        
        if use_gpu:
            
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        
        self.x = x.to(self.device)
        
        self.y = y.to(self.device)
        
        self.names = names
        
        if dimension !=None: 
            
            self.dimension = dimension
            
            self.sis_features = sis_features
            
        else: 
            self.dimension = 3 
            #Maximum terms we will be looking at...
            self.sis_features = 10
            
        #self.dimension = dimension
        self.dimensionality = dimensionality
        
        self.output_dim = output_dim
        #self.sis_features = sis_features
        
        if screening!=None:
            
            self.screening = screening[0]
            
            self.quantile = screening[1]
            
            df_sub = pd.DataFrame(x,columns=names)
            
            screened_space,screened_dimensions = self.feature_space_screening(df_sub,self.dimensionality)
            
            self.x = torch.tensor(screened_space.values)
            
            self.names = screened_space.columns.tolist()
            
            self.dimensionality = screened_dimensions
           
        if self.output_dim!=None:
            
            self.get_dimensions_list()
            
            self.x = self.x[:,self.dimension_less]
            
            x = pd.Series(self.names)
            
            self.names = x.iloc[self.dimension_less].tolist()
            #print(self.names)
            
            

        # Transform the features into standardized format
        self.x_mean = self.x.mean(dim=0)
        
        self.x_std = self.x.std(dim=0)
        
        self.y_mean = self.y.mean()
        
        self.names1 = self.names.copy()
        

        # Transform the target variable value to mean centered
        self.y_centered = self.y - self.y_mean
        
        self.x_standardized = ((self.x - self.x_mean)/self.x_std)
        
        self.scores = []
        
        self.indices = torch.arange(1, (self.dimension*self.sis_features+1)).view(self.dimension*self.sis_features,1).to(self.device)
        
        self.residual = torch.empty(self.y_centered.shape).to(self.device)
        
        self.x_std_clone = torch.clone(self.x_standardized)
        
        
    
    
    def feature_space_screening(self,df_sub,dimensions_screening):
        
        from sklearn.feature_selection import mutual_info_regression

        if self.screening == 'spearman':
            
            spear = spearmanr(df_sub.to_numpy(),self.y,axis=0)
            
            screen1 = abs(spear.statistic)
            
            if screen1.ndim>1:screen1 = screen1[:-1,-1]
            
        elif self.screening=='mi':
            
            screen1 = mutual_info_regression(df_sub.to_numpy(), self.y.numpy())
            
        
        
        df_screening = pd.DataFrame()
        
        df_screening['Feature variables'] = df_sub.columns
        
        df_screening['screen1'] = screen1
        
        df_screening = df_screening.sort_values(by = 'screen1',ascending= False).reset_index(drop=True)
        
        quantile_screen=df_screening.screen1.quantile(self.quantile)
        
        filtered_df = df_screening[(df_screening.screen1 > quantile_screen)].reset_index(drop=True)
        
        if filtered_df.shape[0]==0:
            filtered_df = df_screening[:int(df_sub.shape[1]/2)]

        df_screening1 = df_sub.loc[:,filtered_df['Feature variables'].tolist()]
        
        if len(dimensions_screening) == 0:
            return df_screening1,dimensions_screening
        
        indices = [df_sub.columns.tolist().index(item) for item in df_screening1.columns.tolist() if item in df_sub.columns.tolist()]

        screened_dimensions = np.array(dimensions_screening)[indices]
        
        
        return df_screening1, screened_dimensions
        
        

    '''
    #######################################################################################################

    Constructs the linear equation based on the number of top sis features and the dimension requested.

    #######################################################################################################
    '''
    def higher_dimension(self,iteration):

        #Indices values that needs to be assinged zero, because we already calculated the SIS score of those values..
        ind = (self.indices[:,-1][~torch.isnan(self.indices[:,-1])]).to(self.device)
        
        # Assign them to zero and calculate the score
        self.x_standardized[:,ind.tolist()] = 0

        scores= torch.abs(torch.mm(self.residual,self.x_standardized))
        
        #Assign nan values to zero...

        scores[torch.isnan(scores)] = 0
        
        #Replce the zero values with the original values to compute the combinations and multivariable regression..

        self.x_standardized[:,ind.tolist()] = self.x_std_clone[:,ind.tolist()]

        sorted_scores, sorted_indices = torch.topk(scores, k= self.sis_features)

        sorted_indices = sorted_indices.T
        
        sorted_indices_earlier = self.indices[:((iteration-1)*self.sis_features),(iteration-1)].unsqueeze(1)

        sorted_indices = torch.cat((sorted_indices_earlier,sorted_indices),dim=0)
        
        #Checks for the shape of dimension and sis features needs to be screened and if it is less then adds the remaining count of nan rows..
        if sorted_indices.shape[0] < self.indices.shape[0]:
            
            remaining = (self.sis_features*self.dimension) - int(sorted_indices.shape[0])
            
            nan = torch.full((remaining,1),float('nan')).to(self.device)
            
            sorted_indices = torch.cat((sorted_indices,nan),dim=0)
            
            self.indices = torch.cat((self.indices,sorted_indices),dim=1)

        else:
            self.indices = torch.cat((self.indices,sorted_indices),dim=1)

        comb1 = self.indices[:,-1][~torch.isnan(self.indices[:,-1])]
        
        combinations_generated = torch.combinations(comb1,(int(self.indices.shape[1])-1))
        
        y_centered_clone = self.y_centered.unsqueeze(1).repeat(len(combinations_generated.tolist()),1,1).to(self.device)
        
        comb_tensor = self.x_standardized.T[combinations_generated.tolist(),:]
        
        x_p = comb_tensor.permute(0,2,1)
        
        start_c = time.time()

        sol, _, _, _ = torch.linalg.lstsq(x_p, y_centered_clone)
        
        predicted = torch.matmul(x_p,sol)
        
        residuals = y_centered_clone - predicted
        
        square = torch.square(residuals)
        
        mean = torch.mean(square,dim=1,keepdim=True)
        
        min_value, min_index = torch.min(mean, dim=0)

        coefs_min = torch.squeeze(sol[min_index]).unsqueeze(1)
        
        indices_min  = torch.squeeze(combinations_generated[min_index])
        
        non_std_coeff = ((coefs_min.T/self.x_std[indices_min.tolist()]))
        
        non_std_intercept = self.y.mean() - torch.dot(self.x_mean[indices_min.tolist()]/self.x_std[indices_min.tolist()],coefs_min.flatten())
        
        self.residual = self.y_centered - torch.mm(coefs_min.T,self.x_standardized[:,indices_min.tolist()].T)
        
        rmse = float(torch.sqrt(torch.mean(self.residual**2)))

        r2 = 1 - (float(torch.sum(self.residual**2))/float(torch.sum((self.y_centered)**2)))
        
        terms = []

        for i in range(len(non_std_coeff.squeeze())):
            
            term = str(float(non_std_coeff.squeeze()[i])) + "*" + str(self.names[int(indices_min[i])])
            
            terms.append(term)

        return float(rmse),terms,non_std_intercept,non_std_coeff,r2
    
    
    '''
    ###################################################################
    
    Defines the function to get the dimensionality of feature variables and divide them into separate dimensionality...
    
    ####################################################################
    '''
    
    def get_dimensions_list(self):
        
        #get the same dimensions from the list along with their index position.. 
        result ={}
        
        for index, value in enumerate(self.dimensionality):
            
            if value not in result:
                
                result[value] = []
                
            result[value].append(index)
            
        
        
        if self.output_dim in result.keys():
            
            
            print('************************************************ Extraction of target dimension feature variables found.., performing the regression!!.. ************************************************ \n')
            
            self.dimension_less = result[self.output_dim]
            
            del result[self.output_dim]
            
            print(f'************************************************ {len(self.dimension_less)} output dimension feature variables found in the given list!! ************************************************ \n')
        
            self.dimensions_index_dict = result
            
            del result
            
            
            return self.dimensions_index_dict, self.dimension_less
        
        else:
            
            print('No target dimension feature variables found.. exiting the program..')
            sys.exit()

    '''
    ##########################################################################################################################

    Defines the function to model the equation

    ##########################################################################################################################
    '''
    def regressor_fit(self):
        
        if self.x.shape[1] > self.sis_features*self.dimension :
            print('\n')
            print(f"************************************************ Starting SyMANTIC model in {self.device} ************************************************")
            print('\n')
        else:
            print('************************************************ Given Number of features in SIS screening is greater than the feature space created, changing the SIS features to shape of features created ************************************************ \n')
            
            self.sis_features = int(self.x.shape[1])
            
            self.indices = torch.arange(1, (self.dimension*self.sis_features+1)).view(self.dimension*self.sis_features,1).to(self.device)
            
            
        #Looping over the dimensions
        final_equations=[]
        for i in range(1,self.dimension+1):
            
            if i ==1:
                
                start_1D = time.time()

                #calculate the scores
                scores = torch.abs(torch.mm(self.y_centered.unsqueeze(1).T,self.x_standardized))

                #Set the NaN values claculation to zero, instead of removing 
                scores[torch.isnan(scores)] = 0
                

                #Sort the top number of scores based on the sis_features 
                sorted_scores, sorted_indices = torch.topk(scores,k=self.sis_features)
                
                sorted_indices = sorted_indices.T
                
                remaining = torch.tensor((self.sis_features*self.dimension) - int(sorted_indices.shape[0])).to(self.device)

                #replace the remaining indices with nan
                nan = torch.full((remaining,1),float('nan')).to(self.device)
                
                sorted_indices = torch.cat((sorted_indices,nan),dim=0)
                
                #store the sorted indices as next column
                
                self.indices = torch.cat((self.indices,sorted_indices),dim=1)
                
                selected_index = self.indices[0,1]
                
                x_in = self.x[:, int(selected_index)].unsqueeze(1)

                # Add a column of ones to x for the bias term
                x_with_bias = torch.cat((torch.ones_like(x_in), x_in), dim=1).to(self.device)

                #Calculate the intercept and coefficient, Non standardized
                coef1, _, _, _ = torch.linalg.lstsq(x_with_bias, self.y) #Change to x_with_bias 

                
                #Calculate the residuals based on the standardized and centered values
                x_in1 = self.x_standardized[:, int(selected_index)].unsqueeze(1)

                # Add a column of ones to x for the bias term
                x_with_bias1 = torch.cat((torch.ones_like(x_in1), x_in1), dim=1)
                
                coef, _, _, _ = torch.linalg.lstsq(x_with_bias1, self.y_centered)
            

                self.residual = (self.y_centered - (coef[1]*self.x_standardized[:, int(selected_index)])).unsqueeze(1).T
                rmse = float(torch.sqrt(torch.mean(self.residual**2)))
                
                r2 = 1 - (float(torch.sum(self.residual**2))/float(torch.sum((self.y_centered)**2)))
                #from torcheval.metrics.functional import r2_score
                rmse = float(torch.sqrt(torch.mean(self.residual**2)))
                #self.residual = residual
                coefficient = coef[1]
                
                intercept = self.y.mean() - torch.dot((self.x_mean[int(selected_index)]/self.x_std[int(selected_index)]).reshape(-1), coef[1].reshape(-1))#coef1[0]
                
                if intercept >= 0:
                    coefficient = coef[1]/self.x_std[int(selected_index)]
                    
                    coefficient = "{:.20f}".format(float(coefficient))
                    
                    equation =  coefficient+ '*' + str(self.names[int(selected_index)]) + '+' + str(float(intercept))
                    
                    print('Equation: ', equation)
                    
                    print('RMSE: ', rmse)
                    
                    print('R2:',r2)
                    
                else:
                    
                    coefficient = coef[1]/self.x_std[int(selected_index)]
                    
                    coefficient = "{:.20f}".format(float(coefficient))
                    
                    equation = coefficient + '*' + str(self.names[int(selected_index)])  + str(float(intercept))
                    
                    print('Equation: ', equation)
                    
                    print('RMSE: ', rmse)
                    
                    print('R2::',r2)
                #print('Time taken to generate one dimensional equation: ', time.time()-start_1D,'seconds')
                if self.device == 'cuda':torch.cuda.empty_cache()
                final_equations.append(equation)
                
                
                
            else:
                start = time.time()

                rmse,terms,intercept,coefs,r2 = self.higher_dimension(i)
                equation =''
                for k in range(len(terms)):
                    
                    print(f'{k+1} term in the equation is {terms[k]}')
                    
                    print('\n')

                    if coefs.flatten()[k] > 0:
                        
                        equation = equation + ' + ' + (str(terms[k]))+'  '
                        
                    else:
                        
                        equation = equation + (str(terms[k])) + '  '

                
                
                if intercept>0: equation = equation +'+'+ str(float(intercept))
                
                else: equation = equation + str(float(intercept))

                print('Equation: ',equation[:len(equation)-1])
                print('\n')

                print('RMSE:',float(rmse))
                print('\n')
                
                print('R2::',r2)

                print(f'Time taken for {i} dimension is: ', time.time()-start)
                
                if self.device == 'cuda': torch.cuda.empty_cache()
                final_equations.append(equation)
                
                

        return float(rmse),equation,r2,final_equations


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 17 09:33:17 2023

@author: muthyala.7
"""

import setuptools


setuptools.setup(
    name="TorchSisso",
    version="0.1.5",
    author="Madhav Muthyala",
    author_email="madhavreddymuthyala@gmail.com",
    description="",
    url="https://github.com/PaulsonLab/TorchSISSO",
    packages=setuptools.find_packages(),
    install_requires=[
        "torch",
        "pandas",
        "numpy",
        "scipy",
        "scikit-learn"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 17 09:22:50 2023

@author: muthyala.7
"""

from .FeatureSpaceConstruction import feature_space_construction

from .Regressor import Regressor

from .Regressor_dimension import Regressor 

from .DimensionalFeatureSpaceConstruction import feature_space_construction

from .model import sisso_model



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat May 25 10:17:36 2024

@author: muthyala.7
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 16 12:22:25 2024

@author: muthyala.7
"""

import torch
import pandas as pd
import numpy as np
import warnings
warnings.filterwarnings('ignore')
import time
from itertools import combinations
import sys 
from scipy.stats import spearmanr
import pdb
from sympy import symbols,Pow,powdenest



import math


class feature_space_construction:
    
    '''
    Define the function to get the variables
    '''
    
    def __init__(self,df,operators=None,relational_units = None,initial_screening = None,no_of_operators=None,device='cpu',dimensionality=None,metrics=[0.06,0.995],output_dim=None):
    
      print(f'************************************* Starting Feature Space Construction in {device} ****************************************************************')
      print('\n')
      '''
      ###########################################################################################
    
      no_of_operators - defines the presence of operators (binary or unary) in the expanded features space
    
      For example: if no_of_operators = 2 then the space will be limited to formation of features with 3 operators (x1+x2)/x3 or exp(x1+x2)
    
      ###########################################################################################
      '''
      self.no_of_operators = no_of_operators
    
      self.df = df
      
      '''
      ###########################################################################################
    
      operators [list type]: Defines the mathematical operators needs to be used in the feature expansion
    
      Please look at the README.md for type of mathematical operators allowed
    
      ###########################################################################################
      '''
      if operators !=None: self.operators = operators
      
      else:
          
          print('************************************************ WARNING:: INPUT OPERATOR SET IS NOT PROVIDED, WILL BE USING THE DEFAULT SET OF OPERATORS WHICH MIGHT TAKE MEMORY DEPENDIN THE FEATURE SPACE PROVIDED.. ********************************************** \n')
          self.operators = ['+','-','*','/','exp','sin','cos','tanh','pow(1/2)','pow(1/3)','log','ln','^-1','pow(2)','pow(3)','exp(-1)','/2','+1','-1','/2pi','*2pi']
          
      self.device = torch.device(device)
      
      print(f'************************************ Provided Operator Set to perform Symbolic Regression is:: {self.operators} ************************************************************* \n')
    
      # Filter the dataframe by removing the categorical datatypes and zero variance feature variables
      print('***************************************  Removing the categorical variable columns, if there are any!!  *********************************************************** \n')
      
      self.df = self.df.select_dtypes(include=['float64','int64'])
      
      
      #Checking if we should go for the dimensionality 
      
      self .dimensionality = dimensionality
      
      if len(self.dimensionality) != self.df.shape[1] - 1: sys.exit('Given dimensionality is not matching the number of features given.!!')
      
      if self.dimensionality !=None:
          
          print('**********************************************************  Extracting the dimensions of the same variables and will perform the feature expansion accordingly.... *********************************************************************** \n')
          
          self.relational_units = relational_units
          
    
      # Pop out the Targer variable of the problem and convert to tensor
      self.df.rename(columns = {f'{self.df.columns[0]}':'Target'},inplace=True)
      
      self.Target_column = torch.tensor(self.df.pop('Target')).to(self.device)
      # If initial screening is  yes then do the mic screening.....
      
      if initial_screening != None:
          
          self.screening = initial_screening[0]
          
          self.quantile = initial_screening[1]
          
          self.df, self.dimensionality = self.feature_space_screening(self.df, self.dimensionality)
          
          self.dimensionality = list(self.dimensionality)
          
          
          
    
      # Create the feature values tensor
      self.df_feature_values = torch.tensor(self.df.values).to(self.device)
      
      self.feature_names = self.df.columns.tolist()
      
      self.dimensionality = symbols(self.dimensionality)
      
      self.rmse_metric = metrics[0]
      
      self.r2_metric = metrics[1]
      
      self.metrics = metrics
      
      self.output_dim = output_dim
    



    def get_dimensions_list(self):
        
        # Check for the shape of the feature variables and the length of the provided dimension
        
        if self.df_feature_values.shape[1] == len(self.dimensionality):
            
            print('************************************************ Shape of the dimension list and feature variable count matched... proceeding for further extraction and feature expansion..************************************************ \n')
            
        else:
            
            sys.exit('Mismatch between the dimension list provided and the number of feature variables... \n Please check the dimension list and feature variables and rerun the scipt.. \n ')
        
        #get the same dimensions from the list along with their index position.. 
        result ={}
        
        for index, value in enumerate(self.dimensionality):
            
            if value not in result:
                
                result[value] = []
                
            result[value].append(index)
        
        print('************************************ Extraction of dimensionless and same dimension variables is completed!!.. ********************************************** \n')
        
        if symbols('1') in result.keys():
            
            self.dimension_less = result[symbols('1')]
            
            del result[symbols('1')]
            
            print(f'************************************************ {len(self.dimension_less)} dimension less feature variables found in the given list!! ************************************************ \n')
        
            self.dimensions_index_dict = result
            
            del result
            
            #print(self.dimension_less,self.dimensions_index_dict)
            return self.dimensions_index_dict, self.dimension_less
        
        else:
            
            self.dimensions_index_dict = result
            
            self.dimension_less = None
            
            return self.dimensions_index_dict,self.dimension_less
        
        
    def replace_strings_with_other_elements(self,target_strings,relational_units):

        # Function to find the other element for a single target string
        def find_other_element(target_string):
            
            found_tuple = next((tup for tup in relational_units if target_string in tup), None)
            
            if found_tuple:
                
                return found_tuple[1] if found_tuple[0] == target_string else found_tuple[1]
            
            return target_string  # Return the target string itself if no other element is found

        # List comprehension to replace each target string with its other element
        #print('Replaced the dimensions with the relational units..')
        
        return [find_other_element(target_string) for target_string in target_strings]


    
    
    #### Feature expansion using dimension less numbers....
    
    def dimensionless_feature_expansion(self):
        
        feature_values_non_dimensional = torch.empty(self.df.shape[0],0).to(self.device)
        
        feature_names_non_dimensional =[]
        
        if self.dimension_less is None: 
            
            non_dimensions =[]
            
            print('****************************  Non-Dimension feature expansion is skipped because of no non-dimension features.... ********************************************************** \n ')
            
            return feature_values_non_dimensional,feature_names_non_dimensional,non_dimensions
        
        non_dim_features = self.df_feature_values[:,self.dimension_less]
        
        non_dimensional_features = np.array(self.feature_names)[self.dimension_less]
        
        print('******************************************************  Doing feature expansion of the non dimension feature variables... ******************************************************************** \n')
        
        for op in self.operators:
            
            
            ## Create an empty mutatable tensor and feature names list ####
            transformed_features = torch.empty(self.df.shape[0],0).to(self.device)
            
            transformed_feature_names = []
            
            #Transform the feature variables with exponential mathematical operator
            
            if op == 'exp':
                
                exp = torch.exp(non_dim_features)
                
                transformed_features = torch.cat((transformed_features,exp),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(exp('+ x + "))", non_dimensional_features)))
                
            elif op == '/2':
                
                div2 = non_dim_features/2
                
                transformed_features = torch.cat((transformed_features,div2),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(('+ x + ")/2)", non_dimensional_features)))
                
            elif op == '+1':
                
                div2 = non_dim_features +1
                
                transformed_features = torch.cat((transformed_features,div2),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(('+ x + "+1))", non_dimensional_features)))
                
            elif op == '-1':
                
                div2 = non_dim_features -1
                
                transformed_features = torch.cat((transformed_features,div2),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(('+ x + "-1))", non_dimensional_features)))
            
            elif op == '/2pi':
                
                div2 = non_dim_features/(2*math.pi)
                
                transformed_features = torch.cat((transformed_features,div2),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(('+ x + "/2pi))", non_dimensional_features)))
                
            elif op == '*2pi':
                
                div2 = non_dim_features*(2*math.pi)
                
                transformed_features = torch.cat((transformed_features,div2),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(('+ x + "*2pi))", non_dimensional_features)))

            #Transform the feature variables with natural log mathematical operator
            
            elif op =='ln':
                
                ln = torch.log(non_dim_features)
                
                transformed_features = torch.cat((transformed_features,ln),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(ln('+x + "))", non_dimensional_features)))
                
            
            #Transform the feature variables with log10 mathematical operator
            
            elif op =='log':
                
                log10 = torch.log10(non_dim_features)
                
                transformed_features = torch.cat((transformed_features,log10),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(log('+x + "))", non_dimensional_features)))
                
            elif "pow" in op:
                
                import re
                
                pattern = r'\(([^)]*)\)'
                
                matches = re.findall(pattern, op)
                
                op = eval(matches[0])
                
                transformation = torch.pow(non_dim_features,op)
                
                transformed_features = torch.cat((transformed_features,transformation),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '('+x + f")^{matches[0]}", non_dimensional_features)))

            #Transform the feature variables with SINE mathematical operator
            
            elif op =='sin':
                
                sin = torch.sin(non_dim_features)
                
                transformed_features = torch.cat((transformed_features,sin),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(sin('+x + "))", non_dimensional_features)))
             #Transform the feature variables with COSINE mathematical operator
             
            elif op =='cos':
                
                cos = torch.cos(non_dim_features)
                
                transformed_features = torch.cat((transformed_features,cos),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(cos('+x + "))", non_dimensional_features)))

                
            #Transform the feature variables with reciprocal transformation
            
            elif op =='^-1':
                
                reciprocal = torch.reciprocal(non_dim_features)
                
                transformed_features = torch.cat((transformed_features,reciprocal),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(('+x + ")^(-1))", non_dimensional_features)))
                
            
            #Transform the feature variables with inverse exponential mathematical operator
            
            elif op =='exp(-1)':
                
                exp = torch.exp(non_dim_features)
                
                expreciprocal = torch.reciprocal(exp)
                
                transformed_features = torch.cat((transformed_features,expreciprocal),dim=1)
                
                transformed_feature_names.extend(list(map(lambda x: '(exp('+x + ")^(-1))", non_dimensional_features)))
                
                
            elif op == '+':
                
                if non_dim_features.shape[1] ==1: continue
                
                #generate the combinations on the fly
                combinations1 = list(combinations(non_dimensional_features,2))
                
                combinations2 = torch.combinations(torch.arange(non_dim_features.shape[1]),2)
                
                comb_tensor = non_dim_features.T[combinations2,:]
                
                del combinations2
                
                comb_tensor = comb_tensor.permute(0,2,1)
                
                addition = torch.sum(comb_tensor,dim=2).T
                
                transformed_features = torch.cat((transformed_features,addition),dim=1)
                
                transformed_feature_names.extend(list(map(lambda comb: '('+'+'.join(comb)+')', combinations1)))
                
                del combinations1
                
                
            elif op == '-':
                
                if non_dim_features.shape[1] ==1: continue
                
                combinations1 = list(combinations(non_dimensional_features,2))
                
                combinations2 = torch.combinations(torch.arange(non_dim_features.shape[1]),2)
                
                comb_tensor = non_dim_features.T[combinations2,:]
                
                del combinations2
                
                comb_tensor = comb_tensor.permute(0,2,1)
                
                
                
                sub = torch.sub(comb_tensor[:,:,0],comb_tensor[:,:,1]).T
                
                transformed_features = torch.cat((transformed_features,sub),dim=1)
                
                transformed_feature_names.extend(list(map(lambda comb: '('+'-'.join(comb)+')', combinations1)))
                
                del combinations1
                
            elif op == '*':
                
                if non_dim_features.shape[1] ==1: continue
                
                combinations1 = list(combinations(non_dimensional_features,2))
                
                combinations2 = torch.combinations(torch.arange(non_dim_features.shape[1]),2)
                
                comb_tensor = non_dim_features.T[combinations2,:]
                
                del combinations2
                
                comb_tensor = comb_tensor.permute(0,2,1)

                mul = torch.multiply(comb_tensor[:,:,0],comb_tensor[:,:,1]).T
                
                transformed_features = torch.cat((transformed_features,mul),dim=1)
                
                transformed_feature_names.extend(list(map(lambda comb: '('+'*'.join(comb)+')', combinations1)))
                
                del combinations1
                
            elif op == '/':
                
                if non_dim_features.shape[1] ==1: continue
                
                combinations1 = list(combinations(non_dimensional_features,2))
                
                combinations2 = torch.combinations(torch.arange(non_dim_features.shape[1]),2)
                
                comb_tensor = non_dim_features.T[combinations2,:]
                
                del combinations2
                
                comb_tensor = comb_tensor.permute(0,2,1)
                
                
                
                div1 = torch.div(comb_tensor[:,:,0],comb_tensor[:,:,1]).T
                
                div2 = torch.div(comb_tensor[:,:,1],comb_tensor[:,:,0]).T
                
                transformed_features = torch.cat((transformed_features,div1,div2),dim=1)
                
                transformed_feature_names.extend(list(map(lambda comb: '('+'/'.join(comb)+')', combinations1)))
                
                transformed_feature_names.extend(list(map(lambda comb: '('+'/'.join(comb[::-1])+')', combinations1)))
                
                del combinations1
                
                
            feature_values_non_dimensional = torch.cat((feature_values_non_dimensional,transformed_features),dim=1)
            
            feature_names_non_dimensional.extend(transformed_feature_names)
            
            
            # Check for the list of the features created whether it is empty or not and if it is empty return the empty tensors
            
        if len(feature_names_non_dimensional) == 0:
            
            non_dimensions=[]
            
            return feature_values_non_dimensional,feature_names_non_dimensional,non_dimensions
        
        # if feature names of non dimensional expansion is not zero then check for the nan and inf columns 
        
        else:
            
            nan_columns = torch.any(torch.isnan(feature_values_non_dimensional), dim=0)
            
            inf_columns = torch.any(torch.isinf(feature_values_non_dimensional), dim=0)
            
            nan_inf_columns = nan_columns|inf_columns
            
            feature_values_non_dimensional = feature_values_non_dimensional[:,~nan_inf_columns]
            
            feature_names_non_dimensional = [elem for i,elem in enumerate(feature_names_non_dimensional) if not nan_inf_columns[i]]
            
            non_dimensions = [symbols('1')]*feature_values_non_dimensional.shape[1]
            
            print('************************************************ Completed non dimensional feature expansion with features:', feature_values_non_dimensional.shape[1],'************************************************************** \n')
            
            return feature_values_non_dimensional, feature_names_non_dimensional,non_dimensions
            
            
            
            
    def dimension_to_non_dimension_feature_expansion(self):
        
        print('************************************************ Starting the feature expansion for converting the dimensional to non dimensional features ************************************************ \n')
        
        # We are converting dimensional feature space to non-dimensional feature space
        dim_to_non_dim_feature_values = torch.empty(self.df.shape[0],0).to(self.device)
        
        dim_to_non_dim_feature_names=[]
        
        dim_to_non_dim_units =[]
        
        for dimension, batch in self.dimensions_index_dict.items():

            '''
            will perform the feature expansion converting the dimensional feature spaces to non dimensional feature space by applying operators like 
            
            exp, sin, cos, tan, log, ln, tan, tanh, sinh
            
            '''
            dim_features_values = self.df_feature_values[:,batch]
            
            dim_features_names = np.array(self.feature_names)[batch]
            
            for op in self.operators:
                
                transformed_features = torch.empty(self.df.shape[0],0).to(self.device)
                
                transformed_feature_names = []
                
                if op == 'exp':
                    exp = torch.exp(dim_features_values)
                    
                    transformed_features = torch.cat((transformed_features,exp),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(exp('+ x + "))", dim_features_names)))
   
                #Transform the feature variables with natural log mathematical operator
                
                elif op =='ln':
                    
                    ln = torch.log(dim_features_values)
                    
                    transformed_features = torch.cat((transformed_features,ln),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(ln('+x + "))", dim_features_names)))
                    
                #Transform the feature variables with log10 mathematical operator
                
                elif op =='log':
                    
                    log10 = torch.log10(dim_features_values)
                    
                    transformed_features = torch.cat((transformed_features,log10),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(log('+x + "))", dim_features_names)))
                    
                 
                elif op =='sin':
                    
                    sin = torch.sin(dim_features_values)
                    
                    transformed_features = torch.cat((transformed_features,sin),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(sin('+x + "))", dim_features_names)))
                    
                 #Transform the feature variables with COSINE mathematical operator
                 
                elif op =='cos':
                    
                    cos = torch.cos(dim_features_values)
                    
                    transformed_features = torch.cat((transformed_features,cos),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(cos('+x + "))", dim_features_names)))
                    
                 #Transform the feature variables with inverse exponential mathematical operator
                
                elif op =='exp(-1)':
                    
                    exp = torch.exp(dim_features_values)
                    
                    expreciprocal = torch.reciprocal(exp)
                    
                    transformed_features = torch.cat((transformed_features,expreciprocal),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(exp('+x + ")^(-1))", dim_features_names)))
                    
                    
                dim_to_non_dim_feature_values = torch.cat((dim_to_non_dim_feature_values,transformed_features),dim=1)
                
                dim_to_non_dim_feature_names.extend(transformed_feature_names)
            
            
    
                
        nan_columns = torch.any(torch.isnan(dim_to_non_dim_feature_values), dim=0)
        
        inf_columns = torch.any(torch.isinf(dim_to_non_dim_feature_values), dim=0)
        
        nan_inf_columns = nan_columns|inf_columns
        
        dim_to_non_dim_feature_values = dim_to_non_dim_feature_values[:,~nan_inf_columns]
        
        dim_to_non_dim_feature_names = [elem for i,elem in enumerate(dim_to_non_dim_feature_names) if not nan_inf_columns[i]]
        
        dim_to_non_dim_units = [symbols('1')]*dim_to_non_dim_feature_values.shape[1]
        
        print('************************************************ Dimension to nondimension feature expansion completed.... with feature space size:', dim_to_non_dim_feature_values.shape[1],'************************************************ \n')
        
        
        return dim_to_non_dim_feature_values, dim_to_non_dim_feature_names,dim_to_non_dim_units
        
        
    def dimension_feature_expansion(self,iteration):
        
        dimension_features_values = torch.empty(self.df.shape[0],0).to(self.device)
        
        dimension_features_names =[]
        
        dimension_values = [] #since we can't use tensors for strings we are going to add for the mul and other operators 
        
        
        for dimension,batch in self.dimensions_index_dict.items():
            
            dim_features_values = self.df_feature_values[:,batch]
            
            dim_features_names = np.array(self.feature_names)[batch]
            
            dimension_copy = dimension

            for op in self.operators:
                
                
                transformed_features = torch.empty(self.df.shape[0],0).to(self.device)
                
                transformed_feature_names = []
                
                if op == '+':
                    
                    if len(dim_features_names) == 1: 
                        continue

                    
                    combinations1 = list(combinations(dim_features_names,2))
                    
                    combinations2 = torch.combinations(torch.arange(dim_features_values.shape[1]),2)
                    
                    comb_tensor = dim_features_values.T[combinations2,:]
                    
                    comb_tensor = comb_tensor.permute(0,2,1)
                    
                    addition = torch.sum(comb_tensor,dim=2).T
                    
                    del combinations2
                    #pdb.set_trace()
                    transformed_features = torch.cat((transformed_features,addition),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda comb: '('+'+'.join(comb)+')', combinations1)))
                    
                    del addition,combinations1,comb_tensor
                    
                    dimensions_screened = [dimension]*transformed_features.shape[1]
                    
                    
                    if self.relational_units!=None:
                        
                        dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)

                    dimension_values.extend(dimensions_screened)
                    
                    
                    
                elif op =='-':
                    
                    if len(dim_features_names) == 1: 
                        continue
                    
                    combinations1 = list(combinations(dim_features_names,2))
                    
                    combinations2 = torch.combinations(torch.arange(dim_features_values.shape[1]),2)
                    
                    comb_tensor = dim_features_values.T[combinations2,:]
                    
                    comb_tensor = comb_tensor.permute(0,2,1)
                    
                    sub = torch.sub(comb_tensor[:,:,0],comb_tensor[:,:,1]).T
                    
                    del combinations2
                    
                    transformed_features = torch.cat((transformed_features,sub),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda comb: '('+'-'.join(comb)+')', combinations1)))
                    
                    del combinations1,comb_tensor,sub
                    
                    dimensions_screened = [dimension]*transformed_features.shape[1]
                    
                    if self.relational_units!=None:
                        
                        dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)
                    
                    #add the dimension to the feature variables created
                    dimension_values.extend(dimensions_screened)
                    
                elif op =='*':
                    
                    if len(dim_features_names) == 1: 
                        continue
                    
                    combinations1 = list(combinations(dim_features_names,2))
                    
                    combinations2 = torch.combinations(torch.arange(dim_features_values.shape[1]),2)
                    
                    comb_tensor = dim_features_values.T[combinations2,:]
                    
                    comb_tensor = comb_tensor.permute(0,2,1)
                    
                    mul = torch.multiply(comb_tensor[:,:,0],comb_tensor[:,:,1]).T
                    
                    del combinations2,comb_tensor
                    
                    transformed_features = torch.cat((transformed_features,mul),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda comb: '('+'*'.join(comb)+')', combinations1)))
                    
                    del combinations1,mul
                    
                    dimension = Pow(dimension,2)
                    
                    dimensions_screened = [dimension]*transformed_features.shape[1]
                    
                    if self.relational_units!=None:
                        
                        dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)
                    
                    #add the dimension to the feature variables created
                    dimension_values.extend(dimensions_screened)
                    
                    dimension=dimension_copy
                
                elif "pow" in op:
                    
                    import re
                    
                    pattern = r'\(([^)]*)\)'
                    
                    matches = re.findall(pattern, op)
                    
                    op = eval(matches[0])
                    
                    transformation = torch.pow(dim_features_values,op)
                    
                    transformed_features = torch.cat((transformed_features,transformation),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '('+x + f")^{matches[0]}", dim_features_names)))

                    dimension = powdenest(Pow(dimension,op), force=True)

                    if len(dim_features_names) == 1: 
                        #pdb.set_trace()
                        dimension=[dimension]
                        
                        if self.relational_units!=None:
                            
                            dimension = self.replace_strings_with_other_elements(dimension, self.relational_units)
                            
                        dimension_values.extend(dimension)
                        
                    else:
                        
                        dimensions_screened = [dimension]*transformed_features.shape[1]

                        if self.relational_units!=None:
                        
                            dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)
                        
                        #add the dimension to the feature variables created
                        dimension_values.extend(dimensions_screened)
                        
                    dimension=dimension_copy
                    
                    
                
                elif op =='+1':

                    sum1 = dim_features_values + 1
                    
                    transformed_features = torch.cat((transformed_features,sum1),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(('+x + "+1))", dim_features_names)))
                    
                    if len(dim_features_names) == 1: 
                        
                        dimension=[dimension]
                        
                        if self.relational_units!=None:
                            
                            dimension = self.replace_strings_with_other_elements(dimension, self.relational_units)
                            
                        dimension_values.extend(dimension)
                        
                    else:
                        
                        dimensions_screened = [dimension]*transformed_features.shape[1]

                        if self.relational_units!=None:
                            
                            dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)
                        
                        #add the dimension to the feature variables created
                        dimension_values.extend(dimensions_screened)
                        
                    dimension=dimension_copy
                
                elif op =='-1':

                    sum1 = dim_features_values - 1
                    
                    transformed_features = torch.cat((transformed_features,sum1),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(('+x + "+1))", dim_features_names)))
                    
                    if len(dim_features_names) == 1: 
                        
                        dimension=[dimension]
                        
                        if self.relational_units!=None:
                            
                            dimension = self.replace_strings_with_other_elements(dimension, self.relational_units)
                            
                        dimension_values.extend(dimension)
                        
                    else:
                        
                        dimensions_screened = [dimension]*transformed_features.shape[1]
                        
                        
                        if self.relational_units!=None:
                            
                            dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)
                        
                        #add the dimension to the feature variables created
                        dimension_values.extend(dimensions_screened)
                        
                    dimension=dimension_copy
                    
                elif op =='/2':

                    sum1 = dim_features_values/2
                    
                    transformed_features = torch.cat((transformed_features,sum1),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(('+x + "+1))", dim_features_names)))
                    
                    if len(dim_features_names) == 1: 
                        
                        dimension=[dimension]
                        
                        if self.relational_units!=None:
                            
                            dimension = self.replace_strings_with_other_elements(dimension, self.relational_units)
                            
                        dimension_values.extend(dimension)
                        
                    else:
                        
                        dimensions_screened = [dimension]*transformed_features.shape[1]
                        
                        
                        if self.relational_units!=None:
                            
                            dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)
                        
                        #add the dimension to the feature variables created
                        dimension_values.extend(dimensions_screened)
                        
                    dimension=dimension_copy
                
                elif op =='/2pi':

                    sum1 = dim_features_values/(2*math.pi)
                    
                    transformed_features = torch.cat((transformed_features,sum1),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(('+x + "/2pi))", dim_features_names)))
                    
                    
                    if len(dim_features_names) == 1: 
                        
                        dimension=[dimension]
                        
                        if self.relational_units!=None:
                            
                            dimension = self.replace_strings_with_other_elements(dimension, self.relational_units)
                            
                        dimension_values.extend(dimension)
                        
                    else:
                        
                        dimensions_screened = [dimension]*transformed_features.shape[1]
 
                        if self.relational_units!=None:
                            
                            dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)
                        
                        #add the dimension to the feature variables created
                        dimension_values.extend(dimensions_screened)
                        
                    dimension=dimension_copy
                
                elif op =='*2pi':

                    sum1 = dim_features_values*(2*math.pi)
                    
                    transformed_features = torch.cat((transformed_features,sum1),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(('+x + "*2pi))", dim_features_names)))
                    
                    if len(dim_features_names) == 1: 
                        
                        dimension=[dimension]
                        
                        if self.relational_units!=None:
                            
                            dimension = self.replace_strings_with_other_elements(dimension, self.relational_units)
                            
                        dimension_values.extend(dimension)
                        
                    else:
                        
                        dimensions_screened = [dimension]*transformed_features.shape[1]

                        if self.relational_units!=None:
                            
                            dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)
                        
                        #add the dimension to the feature variables created
                        dimension_values.extend(dimensions_screened)
                        
                    dimension=dimension_copy
  
                    
                elif op =='^-1':

                    inverse = torch.pow(dim_features_values,-1)
                    
                    transformed_features = torch.cat((transformed_features,inverse),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda x: '(('+x + ")^(-1))", dim_features_names)))
                    
                    dimension = Pow(dimension,-1)
                    
                    if len(dim_features_names) == 1: 
                        
                        dimension=[dimension]
                        
                        if self.relational_units!=None:
                            
                            dimension = self.replace_strings_with_other_elements(dimension, self.relational_units)
                            
                        dimension_values.extend(dimension)
                        
                    else:
                        
                        dimensions_screened = [dimension]*transformed_features.shape[1]
                        
                        if self.relational_units!=None:
                            
                            dimensions_screened = self.replace_strings_with_other_elements(dimensions_screened, self.relational_units)
                        
                        #add the dimension to the feature variables created
                        dimension_values.extend(dimensions_screened)
                        
                    dimension=dimension_copy
                    
                
                
                
                
                elif op =='/':
                    
                    if len(dim_features_names) == 1: 
                        continue
                    
                    combinations1 = list(combinations(dim_features_names,2))
                    
                    combinations2 = torch.combinations(torch.arange(dim_features_values.shape[1]),2)
                    
                    comb_tensor = dim_features_values.T[combinations2,:]
                    
                    del combinations2
                    
                    comb_tensor = comb_tensor.permute(0,2,1)
                    
                    div1 = torch.div(comb_tensor[:,:,0],comb_tensor[:,:,1]).T
                    
                    div2 = torch.div(comb_tensor[:,:,1],comb_tensor[:,:,0]).T
                    
                    transformed_features = torch.cat((transformed_features,div1,div2),dim=1)
                    
                    transformed_feature_names.extend(list(map(lambda comb: '('+'/'.join(comb)+')', combinations1)))
                    
                    transformed_feature_names.extend(list(map(lambda comb: '('+'/'.join(comb[::-1])+')', combinations1)))
                    
                    
                    del combinations1,comb_tensor,div1,div2
                    
  
                if op =='/':
                    non_dimensional_div = transformed_features
                    
                    non_dimensional_div_features = transformed_feature_names
                    
                    non_dimensions_units = [symbols('1')]*len(non_dimensional_div_features)
                    
                    nan_columns = torch.any(torch.isnan(non_dimensional_div), dim=0)
                    
                    inf_columns = torch.any(torch.isinf(non_dimensional_div), dim=0)
                    
                    nan_inf_columns = nan_columns|inf_columns
                    
                    non_dimensional_div = non_dimensional_div[:,~nan_inf_columns]
                    
                    non_dimensional_div_features = [elem for i,elem in enumerate(non_dimensional_div_features) if not nan_inf_columns[i]]
                    
                    non_dimensions_units = [elem for i,elem in enumerate(non_dimensions_units) if not nan_inf_columns[i]]
                
                else:
                    
                    #if len(dim_features_names) == 1: 
                    
                    non_dimensional_div = torch.empty(self.df_feature_values.shape[0],0)
                    
                    non_dimensional_div_features = []
                    
                    non_dimensions_units=[]
                    
                    dimension_features_values = torch.cat((dimension_features_values,transformed_features),dim=1)
                    
                    dimension_features_names.extend(transformed_feature_names)
                    #print('operators:',op,'featurenames:',len(dimension_features_names),'feature_values:',dimension_features_values.shape,'dimension_values:',len(dimension_values))
            
        nan_columns = torch.any(torch.isnan(dimension_features_values), dim=0)
        
        inf_columns = torch.any(torch.isinf(dimension_features_values), dim=0)
        
        nan_inf_columns = nan_columns|inf_columns
        
        dimension_features_values = dimension_features_values[:,~nan_inf_columns]
        
        #pdb.set_trace()
        
        dimension_features_names = [elem for i,elem in enumerate(dimension_features_names) if not nan_inf_columns[i]]
        dimension_values = [elem for i,elem in enumerate(dimension_values) if not nan_inf_columns[i]]
        #pdb.set_trace()
        
        
        print('************************************************ Dimensional feature expansion completed.... with feature space size: ',dimension_features_values.shape[1],'************************************************ \n')
        
        if '/' not in self.operators:
            
            non_dimensional_div = torch.empty(self.df_feature_values.shape[0],0)
            
            non_dimensional_div_features = []
            
            non_dimensions_units=[]
        
        try:
            return dimension_features_values, dimension_features_names, dimension_values,non_dimensional_div,non_dimensional_div_features,non_dimensions_units
            
        except:
            
            non_dimensional_div = torch.empty(self.df_feature_values.shape[0],0)
            
            non_dimensional_div_features = []
            
            non_dimensions_units=[]
        
            return dimension_features_values, dimension_features_names, dimension_values,non_dimensional_div,non_dimensional_div_features,non_dimensions_units
    
    
    def inter_dimension_feature_expansion(self,iteration):
        
        combined_batch = torch.empty(self.df.shape[0],0)
        
        combined_dimensions =[]
        
        combined_feature_names=[]
        
        for dimension,batch in self.dimensions_index_dict.items():
            
            combined_batch = torch.cat((combined_batch,self.df_feature_values[:,batch]),dim=1)
            
            combined_dimensions.extend([dimension]*len(batch))
            
            combined_feature_names.extend(np.array(self.feature_names)[batch])
            
        if self.dimension_less !=None:    
            combined_batch = torch.cat((combined_batch,self.df_feature_values[:,self.dimension_less]),dim=1)
                
            combined_dimensions.extend([1]*len(self.dimension_less))
                
            combined_feature_names.extend(np.array(self.feature_names)[self.dimension_less])
        
        
        # do the combinations and perform the multiplication and the division operations
        transformed_features = torch.empty(self.df.shape[0],0)
        
        transformed_feature_names=[]
        
        transformed_dimensions=[]
        
        if '*' in self.operators:
            
            combinations1 = list(combinations(combined_feature_names,2))
            
            combinations2 = torch.combinations(torch.arange(combined_batch.shape[1]),2)
            
            comb_tensor = combined_batch.T[combinations2,:]
            
            comb_tensor = comb_tensor.permute(0,2,1)
            
            mul = torch.multiply(comb_tensor[:,:,0],comb_tensor[:,:,1]).T
            
            del combinations2,comb_tensor
            
            transformed_features = torch.cat((transformed_features,mul),dim=1)
            
            transformed_feature_names.extend(list(map(lambda comb: '('+'*'.join(comb)+')', combinations1)))
            
            del combinations1,mul
            
            combinations1 = list(combinations(combined_dimensions,2))
            
            process_tuple = lambda x, y: (Pow(x,2) if x == y 
                                          else x*y)
            transformed_dimensions.extend(list(map(lambda t: process_tuple(*t), combinations1)))
            
            del combinations1
        
        
        if '/' in self.operators:
            
            combinations1 = list(combinations(combined_feature_names,2))
            
            combinations2 = torch.combinations(torch.arange(combined_batch.shape[1]),2)
            
            comb_tensor = combined_batch.T[combinations2,:]
            
            del combinations2
            
            comb_tensor = comb_tensor.permute(0,2,1)

            div1 = torch.div(comb_tensor[:,:,0],comb_tensor[:,:,1]).T
            
            div2 = torch.div(comb_tensor[:,:,1],comb_tensor[:,:,0]).T
            
            transformed_features = torch.cat((transformed_features,div1,div2),dim=1)
            
            transformed_feature_names.extend(list(map(lambda comb: '('+'/'.join(comb)+')', combinations1)))
            
            transformed_feature_names.extend(list(map(lambda comb: '('+'/'.join(comb[::-1])+')', combinations1)))
            
            del combinations1
            
            combinations1 = list(combinations(combined_dimensions,2))
            
            dimensions=[]
            
            dimensions1=[]
            
            for x,y in combinations1:
                
                if x ==y: 
                    dimensions.append(symbols('1'))
                    
                    dimensions1.append(symbols('1'))
                else: 
                    dimensions.append(x/y)
                    
                    dimensions1.append(y/x)
    
            
            transformed_dimensions.extend(dimensions+dimensions1)
            
            
            del combinations1,dimensions,dimensions1
            
        screened_dimensions = transformed_dimensions
        
        if self.relational_units!=None:
            
            screened_dimensions = self.replace_strings_with_other_elements(screened_dimensions, self.relational_units)
           
        
        transformed_dimensions = screened_dimensions
        
        
        print('************************************************ Inter dimensional feature expansion completed, with feature space size: ', transformed_features.shape[1],'************************************************ \n')
        
        return transformed_features,transformed_feature_names,transformed_dimensions
        
    
    def feature_space_screening(self,df_sub,dimensions_screening):
        
        from sklearn.feature_selection import mutual_info_regression

        if self.screening == 'spearman':
            
            spear = spearmanr(df_sub.to_numpy(),self.Target_column,axis=0)
            
            screen1 = abs(spear.statistic)
            
            if screen1.ndim>1:screen1 = screen1[:-1,-1]
            
        elif self.screening=='mi':
            
            screen1 = mutual_info_regression(df_sub.to_numpy(), self.Target_column.numpy())
            
        
        
        df_screening = pd.DataFrame()
        
        df_screening['Feature variables'] = df_sub.columns
        
        df_screening['screen1'] = screen1
        
        df_screening = df_screening.sort_values(by = 'screen1',ascending= False).reset_index(drop=True)
        
        quantile_screen=df_screening.screen1.quantile(self.quantile)
        
        filtered_df = df_screening[(df_screening.screen1 > quantile_screen)].reset_index(drop=True)
        
        if filtered_df.shape[0]==0:
            filtered_df = df_screening[:int(df_sub.shape[1]/2)]

        df_screening1 = df_sub.loc[:,filtered_df['Feature variables'].tolist()]
        
        if len(dimensions_screening) == 0:
            return df_screening1,dimensions_screening
        
        indices = [df_sub.columns.tolist().index(item) for item in df_screening1.columns.tolist() if item in df_sub.columns.tolist()]

        screened_dimensions = np.array(dimensions_screening)[indices]
        
        
        return df_screening1, screened_dimensions
    
    
    def feature_expansion(self):
        
        
        print('\nStarting the feature expansion for the given feature variables.... \n ')
        
        for i in range(1,self.no_of_operators):
        
            start_time = time.time()
            
            # Get the dimension and non dimension variables... 
            
            self.get_dimensions_list()
            
            #Get the non dimension expansion... 
            
            non_dimension_feature_values, non_dimension_feature_names, non_dimension_units = self.dimensionless_feature_expansion()
            
            # Transform the dimension to non-dimension featur expansion....
            
            dim_to_non_dim_values,dim_to_non_dim_names,dim_to_non_dim_units = self.dimension_to_non_dimension_feature_expansion()
            
            
            #Transform the dimensional feature expansion... 
            
            dim_exp_feature_values, dim_exp_feature_names, dim_exp_units,non_dim_expanded_values,non_dim_expanded_names,non_dim_expanded_units = self.dimension_feature_expansion(i)
            
            # Inter-dimension feature expansion
            
            dim_inter_exp_values, dim_inter_exp_names,dim_inter_exp_units = self.inter_dimension_feature_expansion(i)
            
            # Concatenate the values to feature values, variables, dimensionality......
           
            self.df_feature_values = torch.cat((self.df_feature_values,non_dimension_feature_values,dim_to_non_dim_values,dim_exp_feature_values,dim_inter_exp_values,non_dim_expanded_values),dim=1)
        
            self.feature_names.extend(non_dimension_feature_names + dim_to_non_dim_names + dim_exp_feature_names + dim_inter_exp_names+non_dim_expanded_names)
            
            self.dimensionality.extend(list(non_dimension_units) + list(dim_to_non_dim_units) + list(dim_exp_units) + list(dim_inter_exp_units) + list(non_dim_expanded_units))
            

            end_time = time.time()
            
            print(f'************************************************ Time taken for the {i} feature expansion: ',end_time - start_time, ' seconds.. ************************************************ \n')
            
            print(f'************************************************ Size of the feature space formed in the {i} expansion',self.df_feature_values.shape[1],'************************************************ \n')
            
            
        return self.df_feature_values,self.Target_column,self.feature_names,self.dimensionality
            
        
        

    
        
    




       
     

